// x86obf - a code virtualizer/obfuscator for 32bit PE files
// Copyright (c) 2015 Tomislav Zubcic - chaplja.blogspot.com / x86obf@gmail.com
// $Id: xl_blocks.cpp 652 2015-02-24 21:15:39Z tzubcic $

#include "xl_local.h"

void xl_locate_code_blocks(void)
{
	UINT8		*start, *end, *m_start, *m_end;

	start = xl_offset_to_ptr(xl.code_off);
	end = start + xl.code_size - XM_START_SIZE - XM_END_SIZE;

	if (end <= start) {
		xl_error("Invalid code section found");
	}

	while (start < end) {
		// start marker
		if (!memcmp(start, XM_START, XM_START_SIZE - 1)) {
			m_start = start;
			m_end = NULL;

			start++;

			while (start < end) {
				if (!memcmp(start, XM_START, XM_START_SIZE - 1)) {
					xl_error("Found two start markers in a row, missing end marker (0x%08X)", xl_ptr_to_va(m_start));
				}

				if (!memcmp(start, XM_END, XM_END_SIZE)) {
					m_end = start + XM_END_SIZE;
					break;
				}

				start++;
			}

			if (!m_end) {
				xl_error("Start marker found without a matching end marker (0x%08X)", xl_ptr_to_va(start));
			}

			xl_add_block(m_start, m_end);

			start += XM_END_SIZE;

			continue;
		} else if (!memcmp(start, XM_END, XM_END_SIZE)) {
			xl_error("End marker found without a matching start marker (0x%08X)", xl_ptr_to_va(start));
		}

		start++;
	}
}

void xl_add_block(UINT8 *start, UINT8 *end)
{
	xl_block_t		b;

	b.start = start;
	b.len = end - start;
	b.rva = xl_ptr_to_rva(start);
	b.va = xl_ptr_to_va(start);
	b.end_va = b.va + b.len;

	b.code_start = start + XM_START_SIZE;
	b.code_len = b.len - XM_START_SIZE - XM_END_SIZE;
	b.code_rva = xl_ptr_to_rva(b.code_start);
	b.code_va = xl_ptr_to_va(b.code_start);

	xl.blocks.push_back(b);
}

void xl_iterate_blocks(xl_block_iterate_t callback)
{
	for (unsigned i = 0; i < xl.blocks.size(); i++) {
		callback(&xl.blocks[i]);
	}
}

void xl_disassemble_block(xl_block_t *b)
{
	ud_t		ud;
	UINT32		td, db;

	b->disasm.clear();

	ud_init(&ud);

	ud_set_input_buffer(&ud, b->code_start, b->code_len);
	ud_set_mode(&ud, 32);
	ud_set_syntax(&ud, UD_SYN_INTEL);
	ud_set_pc(&ud, b->code_va);

	td = 0;

	while (ud_disassemble(&ud)) {
		db = ud_insn_len(&ud);

		xl_add_disasm(b, &ud);

		td += db;
	}

	if (td != b->code_len) {
		xl_error("Disassembly failure, unknown instruction");
	}
}

void xl_update_disasm_owner_block(xl_block_t *b)
{
	for (unsigned i = 0; i < b->disasm.size(); i++) {
		b->disasm[i].block = b;

		xl_update_operands_owner_disasm(&b->disasm[i]);
	}

	xl_update_links_disasm(b);
}

void xl_analyze_disassembly_in_block(xl_block_t *b)
{
	xl_analyze_disassembly(b->disasm);
}

void xl_analyze_relocation(xl_block_t *b)
{
	xl_disasm_t			*dis;
	xl_op_t				*o;
	UINT32				rlc_va;

	if (!xl.has_relocs) {
		return;
	}

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		dis = &b->disasm[i];

		for (unsigned j = 0; j < xl.input_relocs.size(); j++) {
			rlc_va = xl.input_relocs[j];

			if (rlc_va >= dis->va && rlc_va < (dis->va + dis->len)) {
				dis->reloc_count++;

				// validate number of operands
				if (!dis->op.size() || dis->op.size() > 2) {
					xl_error("Relocation parsing error");
				}

				// instruction size check
				if (dis->len < ((4 * dis->reloc_count) + 1)) {
					xl_error("Relocation parsing error");
				}

				// validate relocation position
				if (rlc_va > (dis->va + dis->len - 4)) {
					xl_error("Relocation parsing error");
				}

				if (dis->op.size() == 1) {
					o = &dis->op[0];
				} else if (dis->op.size() == 2) {
					if (dis->op[0].type == UD_OP_REG) {
						if (dis->op[1].type == UD_OP_IMM) {
							o = &dis->op[1];
						} else if (dis->op[1].type == UD_OP_MEM) {
							o = &dis->op[1];
						} else {
							xl_code_gen_error(XL_ERR_RELOC_FINDING_BAD_OP);
						}
					} else if (dis->op[0].type == UD_OP_MEM) {
						if (dis->op[1].type == UD_OP_REG) {
							o = &dis->op[0];
						} else if (dis->op[1].type == UD_OP_IMM) {
							if (rlc_va == (dis->va + dis->len - 4)) {
								o = &dis->op[1];
							} else {
								o = &dis->op[0];
							}
						} else {
							xl_code_gen_error(XL_ERR_RELOC_FINDING_BAD_OP);
						}
					} else {
						xl_code_gen_error(XL_ERR_RELOC_FINDING_BAD_OP);
					}
				} else {
					xl_error("Relocation parsing error");
				}

				o->is_reloc = true;
				o->reloc_handled = false;

				o->reloc_offset = rlc_va - o->dis->va;
			}
		}
	}
}

void xl_abort_if_instructions_not_analyzed(xl_block_t *b)
{
	for (unsigned i = 0; i < b->disasm.size(); i++) {
		if (!b->disasm[i].is_analyzed) {
			xl_error("Instruction not analyzed (%s)", b->disasm[i].txt.c_str());
		}
	}
}

void xl_print_block(xl_block_t *b)
{
	printf("Code block at 0x%08X RVA(0x%08X)\n", b->code_va, b->code_rva);

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		printf("[%s] [0x%08X] %s\n", b->disasm[i].reloc_count ? "RLC" : "   ", b->disasm[i].va, b->disasm[i].txt.c_str());
	}
}

bool xl_addr_in_block(xl_block_t *b, UINT32 addr)
{
	if (addr < b->code_va) {
		return false;
	}

	if (addr >= (b->code_va + b->code_len)) {
		return false;
	}

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		if (addr == b->disasm[i].va) {
			return true;
		}
	}

	xl_code_gen_error(XL_ERR_IS_ADDR_IN_BLOCK_CANNOT_LINK);

	return false;
}
