// x86obf - a code virtualizer/obfuscator for 32bit PE files
// Copyright (c) 2015 Tomislav Zubcic - chaplja.blogspot.com / x86obf@gmail.com
// $Id: xl_interface.cpp 652 2015-02-24 21:15:39Z tzubcic $

#include "xl_local.h"

#pragma comment(lib, "shlwapi.lib")

void xl_parse_cmd_line(int argc, char *argv[])
{
	int			c;

	// default values
	xl.opt_num_vms = 0;
	xl.opt_print_blocks = false;

	// parse arguments
	while ((c = getopt(argc, argv, "bi:o:v:")) != EOF) {
		switch (c) {
			case 'i':
				xl.in_path = optarg;
				break;
			case 'o':
				xl.out_path = optarg;
				break;
			case 'v':
				xl.opt_num_vms = strtoul(optarg, NULL, 10);
				break;
			case 'b':
				xl.opt_print_blocks = true;
				break;
			case 'h':
				xl_usage();
				break;
			case '?':
				xl_usage();
				break;
			default:
				xl_usage();
				break;
		}
	}

	// verify command line options
	if (xl.in_path.length() == 0) {
		xl_usage();
	}

	if (xl.out_path.length() == 0) {
		xl_usage();
	}

	if (xl.opt_num_vms == 0) {
		xl.opt_num_vms = 1;
	} else if (xl.opt_num_vms > X_VM_MAX_AMOUNT) {
		printf("WARNING: Maximum number of virtual machines can be %u\n", X_VM_MAX_AMOUNT);

		xl.opt_num_vms = X_VM_MAX_AMOUNT;
	}
}

void xl_load_input_file(void)
{
	// open the input file
	ifstream input_file(xl.in_path, ios::in | ios::binary);

	// check if file exists
	if (!input_file || !input_file.is_open()) {
		xl_error("Cannot open input file: %s\n", xl.in_path.c_str());
	}

	try {
		xl.pe_input = new pe_base(pe_factory::create_pe(input_file));

		xl.pe_type = xl.pe_input->get_pe_type();

		if (xl.pe_type != pe_type_32) {
			xl_error("Unsupported PE type, only 32 bit PE files are supported");
		}

		input_file.seekg(0);

		xl.raw_file.clear();
		xl.raw_len = 0;

		char		data;

		while (input_file.get(data)) {
			xl.raw_file.push_back((UINT8) data);
			xl.raw_len++;
		}

		// should we take care of relocations
		xl.has_relocs = xl.pe_input->has_reloc();

		// original entry point
		xl.oep_rva = xl.pe_input->get_ep();
		xl.oep_va = xl_rva_to_va(xl.oep_rva);

		// module base
		xl.pe_input->get_image_base(*((uint32_t *) &xl.module_base));

		// create a new section
		section new_section;

		new_section.readable(true).writeable(true).executable(true);
		new_section.set_characteristics(new_section.get_characteristics() | 0x20);
		new_section.set_name(".86obf");
		new_section.set_raw_data("If you know how many cupcakes I'm holding behind my back I'll give you both of them.");

		xl.new_section = &xl.pe_input->add_section(new_section);
		xl.pe_input->set_section_virtual_size(*xl.new_section, xl.new_section->get_raw_data().length());

		// new section information
		xl.nsc_rva = xl.new_section->get_virtual_address();
		xl.nsc_va = xl.module_base + xl.nsc_rva;

		// empty new section data
		xl.nsc_data.clear();

		// code section information
		xl.code_rva = xl.pe_input->get_base_of_code();
		xl.code_va = xl_rva_to_va(xl.code_rva);
		xl.code_off = xl_rva_to_offset(xl.code_rva);
		xl.code_data = xl_offset_to_ptr(xl.code_off);
		xl.code_section = &xl.pe_input->section_from_file_offset(xl.code_off);
		xl.code_size = xl.code_section->get_size_of_raw_data();

		// get imports
		xl_get_import_addresses();

		// parse relocation section
		x_parse_input_relocations();
	} catch(const pe_exception& e) {
		xl_error("%s", e.what());
	}
}

void xl_process_input_file(void)
{
	// initialize data management
	xl_init_data_management();

	// register common data
	xl_register_common_data();

	printf("[+] Locating marked code blocks\n");

	// find marked areas
	xl_locate_code_blocks();

	// abort if protection markers haven't been inserted
	if (!xl.blocks.size()) {
		xl_error("No code blocks have been marked for protection");
	}

	// demo limitation
#ifdef XDEMO
	if (xl.blocks.size() > 1) {
		xl_error("You can virtualize only 1 block of code in demo version");
	}
#endif

	printf(" > Found %u code block(s)\n", xl.blocks.size());
	printf("[+] Analyzing code\n");

	// disassemble code blocks
	xl_iterate_blocks(xl_disassemble_block);

	// update all links in blocks
	xl_iterate_blocks(xl_update_disasm_owner_block);

	// analyze disassembly
	xl_iterate_blocks(xl_analyze_disassembly_in_block);

	// analyze relocations
	xl_iterate_blocks(xl_analyze_relocation);

	// make sure all instructions have been analyzed
	xl_iterate_blocks(xl_abort_if_instructions_not_analyzed);

	// output detected code blocks
	if (xl.opt_print_blocks) {
		xl_iterate_blocks(xl_print_block);
	}

	// generate virtual machines
	xl_generate_virtual_machines();

	printf("[+] Virtualizing marked code blocks\n");

	// analyze blocks for virtualization
	xl_iterate_blocks(xl_analyze_block_for_virtualization);

	// assign random vm to blocks
	xl_assign_vm_to_blocks();

	// split blocks
	xl_split_blocks();

	// process and virtualize block parts
	xl_process_block_parts();

	// assemble vm bytecode
	xl_assemble_vm_code();

	// initialize vm data
	xl_set_vm_data();

	// get bytecode entry points
	xl_set_bytecode_entry_points();

	printf("[+] Generating VM code\n");

	// gather information needed to generate VM code
	xl_construct_context_data();

	// process context data
	xl_process_context();

	// parse context reply data
	xl_parse_context_reply_data();

	// write jumps to vm
	xl_assemble_jumps_to_vm();

	// update code section
	xl_update_code_section();

	// write new section data
	xl_assemble_new_section();

	// add relocations if necessary
	xl_add_relocation_section();
}

void xl_save_output_file(void)
{
	auto_ptr<pe_base>		new_image;

	printf("[+] Building output file\n");

	try {
		// create a new pe object
		new_image.reset(static_cast<pe_base*>(new pe_base(pe_properties_32(), xl.pe_input->get_section_alignment())));

		// basic information
		new_image->set_characteristics(xl.pe_input->get_characteristics());
		new_image->set_dll_characteristics(xl.pe_input->get_dll_characteristics());
		new_image->set_file_alignment(xl.pe_input->get_file_alignment());
		new_image->set_heap_size_commit(xl.pe_input->get_heap_size_commit_64());
		new_image->set_heap_size_reserve(xl.pe_input->get_heap_size_reserve_64());
		new_image->set_stack_size_commit(xl.pe_input->get_stack_size_commit_64());
		new_image->set_stack_size_reserve(xl.pe_input->get_stack_size_reserve_64());
		new_image->set_image_base_64(xl.pe_input->get_image_base_64());
		new_image->set_ep(xl.pe_input->get_ep());
		new_image->set_number_of_rvas_and_sizes(new_image->get_number_of_rvas_and_sizes());
		new_image->set_subsystem(xl.pe_input->get_subsystem());

		// add directories
		for (unsigned long i = 0; i < xl.pe_input->get_number_of_rvas_and_sizes(); i++) {
			new_image->set_directory_rva(i, xl.pe_input->get_directory_rva(i));
			new_image->set_directory_size(i, xl.pe_input->get_directory_size(i));
		}

		// sections
		const section_list& pe_sections = xl.pe_input->get_image_sections();

		for (section_list::const_iterator it = pe_sections.begin(); it != pe_sections.end(); ++it) {
			new_image->set_section_virtual_size(new_image->add_section(*it), (*it).get_virtual_size());
		}

		// new entry point
		//new_image->set_ep(xl.new_oep_rva);

		// calculate checksum
		stringstream temp_pe(ios::out | ios::in | ios::binary);
		rebuild_pe(*new_image, temp_pe);
		new_image->set_checksum(calculate_checksum(temp_pe));

		// rebuild the file
		ofstream new_pe_file(xl.out_path, ios::out | ios::binary | ios::trunc);

		if (!new_pe_file) {
			xl_error("Cannot open output file");
		}

		rebuild_pe(*new_image, new_pe_file);

		new_pe_file.close();

		printf("[+] Finished!\n");
	} catch(const pe_exception& e) {
		xl_error("%s", e.what());
	}
}

void xl_usage(void)
{
	puts("Mandatory parameters:");
	puts("  -i <input filename>");
	puts("  -o <output filename>");
	puts("Optional parameters:");
	puts("  -v <number> : number of virtual machines");
	puts("  -b : list marked code blocks");
	puts("  -h : show this help");

	exit(1);
}

void xl_get_directory(void)
{
	if (!GetModuleFileNameA(NULL, xl.cur_dir, sizeof(xl.cur_dir))) {
		xl_code_gen_error(XL_ERR_CANT_GET_LOCAL_DIR);
	}

	PathRemoveFileSpecA(xl.cur_dir);
}
