// x86obf - a code virtualizer/obfuscator for 32bit PE files
// Copyright (c) 2015 Tomislav Zubcic - chaplja.blogspot.com / x86obf@gmail.com
// $Id: x_blocks.cpp 657 2015-02-25 11:42:33Z tzubcic $

#include "x_local.h"

void x_set_block(UINT8 *start, UINT8 *end)
{
	x_block_t		*b;

	b = &xg.ctx->block;

	// clear the disasm list
	b->disasm.clear();

	// set basic information
	b->start = start;
	b->len = end - start;
	b->rva = xg.ctx->new_code_rva;
	b->va = xg.ctx->new_code_va;
	b->end_va = b->va + b->len;

	b->code_start = b->start;
	b->code_len = b->len;
	b->code_va = b->va;
	b->code_rva = b->rva;
}

void x_disassemble_block(void)
{
	x_block_t	*b;
	ud_t		ud;
	UINT32		td, db;

	b = &xg.ctx->block;

	b->disasm.clear();

	ud_init(&ud);

	ud_set_input_buffer(&ud, b->code_start, b->code_len);
	ud_set_mode(&ud, 32);
	ud_set_syntax(&ud, UD_SYN_INTEL);
	ud_set_pc(&ud, b->code_va);

	td = 0;

	while (ud_disassemble(&ud)) {
		db = ud_insn_len(&ud);

		x_add_disasm(b, &ud);

		td += db;
	}

	if (td != b->code_len) {
		x_exit_thread(X_ERR_DISASM_BLOCK_BAD_LENGTH);
	}
}

void x_update_disasm_owner_block(void)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		b->disasm[i].block = b;

		x_update_operands_owner_disasm(&b->disasm[i]);
	}

	x_update_links_disasm(b);
}

void x_analyze_disassembly_in_block(void)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	// parse instruction operands
	for (unsigned i = 0; i < b->disasm.size(); i++) {
		x_add_operands(&b->disasm[i]);
	}

	// analyze instruction and operands
	for (unsigned i = 0; i < b->disasm.size(); i++) {
		x_check_instruction(&b->disasm[i]);
	}
}

void x_analyze_relocation_for_mutation(void)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	x_reloc_info_t		*ri;
	x_reloc_t			*r;
	x_disasm_t			*dis;
	x_op_t				*o;
	UINT32				*v;

	if (!xg.ctx->has_relocs) {
		return;
	}

	ri = &xg.ctx->gen_reloc;

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		dis = &b->disasm[i];

		for (unsigned j = 0; j < ri->relocs.size(); j++) {
			r = &ri->relocs[j];

			if ((r->addr_start >= dis->rva) && (r->addr_end == (dis->rva + dis->len))) {
				dis->reloc_count++;

				// validate number of operands
				if (!dis->op.size() || (dis->op.size() > 2)) {
					x_exit_thread(X_ERR_ANALYZE_RELOCS_PARSING_ERROR);
				}

				// instruction size check
				if (dis->len < ((4 * dis->reloc_count) + 1)) {
					x_exit_thread(X_ERR_ANALYZE_RELOCS_PARSING_ERROR);
				}

				// find the value in arguments
				for (unsigned k = 0; k < dis->op.size(); k++) {
					o = &dis->op[k];

					if (o->type == UD_OP_IMM || o->type == UD_OP_JIMM) {
						v = &o->data.val;
					} else if (o->type == UD_OP_MEM) {
						v = &o->data.mem.offset;
					} else {
						continue;
					}

					if (*v == r->value) {
						o->is_reloc = true;
						o->reloc_handled = false;
						r->found = true;

						break;
					}
				}

				if (!r->found) {
					x_exit_thread(X_ERR_ANALYZE_RELOCS_NOT_FOUND);
				}
			}
		}
	}

	for (unsigned i = 0; i < xg.ctx->gen_reloc.relocs.size(); i++) {
		if (!xg.ctx->gen_reloc.relocs[i].found) {
			x_exit_thread(X_ERR_ANALYZE_RELOCS_NOT_HANDLED);
		}
	}
}

void x_abort_if_instructions_not_analyzed(void)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		if (!b->disasm[i].is_analyzed) {
			x_exit_thread(X_ERR_INS_NOT_ANALYZED);
		}
	}
}

void x_clear_block(void)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		b->disasm[i].op.clear();
	}

	b->disasm.clear();
}

Label *x_addr_to_label(UINT32 addr)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	if (addr < b->code_va) {
		return NULL;
	}

	if (addr >= (b->code_va + b->code_len)) {
		return NULL;
	}

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		if (addr == b->disasm[i].va) {
			return b->disasm[i].label;
		}
	}

	x_exit_thread(X_ERR_ADDR_TO_LABEL_CANNOT_LINK);

	return NULL;
}

bool x_addr_in_block(UINT32 addr)
{
	x_block_t	*b;

	b = &xg.ctx->block;

	if (addr < b->code_va) {
		return false;
	}

	if (addr >= (b->code_va + b->code_len)) {
		return false;
	}

	for (unsigned i = 0; i < b->disasm.size(); i++) {
		if (addr == b->disasm[i].va) {
			return true;
		}
	}

	x_exit_thread(X_ERR_ADDR_IN_BLOCK_CANNOT_LINK);

	return false;
}
