// x86obf - a code virtualizer/obfuscator for 32bit PE files
// Copyright (c) 2015 Tomislav Zubcic - chaplja.blogspot.com / x86obf@gmail.com
// $Id: x_context.cpp 660 2015-02-25 12:11:30Z tzubcic $

#include "x_local.h"

// assembler
#define A			(*xg.code_gen)

void x_initialize_context(void *data, unsigned data_len)
{
	// by default invalid assembler
	xg.code_gen = NULL;
	xg.code_gen_runtime = NULL;
	xg.temp_gen = NULL;
	xg.temp_runtime = 0;

	// context data
	xg.ctx_data = data;

	// context data length
	xg.ctx_data_len = data_len;

	// initial offset
	xg.ctx_data_offset = 0;

	// allocate a new ctx object
	xg.ctx = new x_ctx_t;
}

bool x_parse_context_data(void)
{
	UINT32			protocol;
	UINT32			data_len;

	// make sure minimum amount of data is generated
	if (xg.ctx_data_len < 8) {
		return false;
	}

	// read and validate context protocol
	if (!x_ctx_read_dword(&protocol, false)) {
		return false;
	}

	if (protocol != X_CTX_DATA_PROTOCOL) {
		return false;
	}

	// expected data length
	if (!x_ctx_read_dword(&data_len, false)) {
		return false;
	}

	if (data_len != xg.ctx_data_len) {
		return false;
	}

	// default load address
	if (!x_ctx_read_dword(&xg.ctx->module_base)) {
		return false;
	}

	// relocation support
	if (!x_ctx_read_bool(&xg.ctx->has_relocs)) {
		return false;
	}

	// oep
	if (!x_ctx_read_dword(&xg.ctx->oep_rva)) {
		return false;
	}

	xg.ctx->oep_va = xg.ctx->module_base + xg.ctx->oep_rva;

	// new section rva
	if (!x_ctx_read_dword(&xg.ctx->nsc_rva)) {
		return false;
	}

	xg.ctx->nsc_va = xg.ctx->module_base + xg.ctx->nsc_rva;

	// new section data size
	if (!x_ctx_read_dword(&xg.ctx->nsc_data_size)) {
		return false;
	}

	xg.ctx->new_code_rva = xg.ctx->nsc_rva + xg.ctx->nsc_data_size;
	xg.ctx->new_code_va = xg.ctx->module_base + xg.ctx->new_code_rva;

	// imports
	if (!x_parse_context_data_imports()) {
		return false;
	}

	// registered data addresses
	if (!x_parse_context_data_data_addresses()) {
		return false;
	}

	// vm architectures
	if (!x_parse_context_data_vm_architectures()) {
		return false;
	}

	// block information
	if (!x_parse_context_data_block_information()) {
		return false;
	}

	// make sure all data has been read
	if (xg.ctx_data_offset != xg.ctx_data_len) {
		return false;
	}

	return true;
}

bool x_parse_context_data_vm_architectures(void)
{
	x_vm_t		vm;

	// number of virtual machines
	if (!x_ctx_read_dword(&xg.ctx->num_vms)) {
		return false;
	}

	if (!xg.ctx->num_vms) {
		x_exit_thread(X_ERR_PARSE_CTX_VM_ARCH_TOO_NO_VM);
	}

	if (xg.ctx->num_vms > X_VM_MAX_AMOUNT) {
		x_exit_thread(X_ERR_PARSE_CTX_VM_ARCH_TOO_MANY_VM);
	}

	for (unsigned i = 0; i < xg.ctx->num_vms; i++) {
		// instruction ids
		for (unsigned j = 0; j < X_VMI_TOTAL; j++) {
			if (!x_ctx_read_word(&vm.ins_id[j])) {
				return false;
			}
		}

		// register ids
		for (unsigned j = 0; j < X_VMR_TOTAL; j++) {
			if (!x_ctx_read_word(&vm.reg_id[j])) {
				return false;
			}
		}

		// longest instruction length
		if (!x_ctx_read_dword(&vm.longest_instruction)) {
			return false;
		}

		// data addresses
		if (!x_ctx_read_dword(&vm.data_rva_vmi_rvas)) {
			return false;
		}

		if (!x_ctx_read_dword(&vm.data_rva_vmi_lengths)) {
			return false;
		}

		if (!x_ctx_read_dword(&vm.data_va_tls_slot)) {
			return false;
		}
		
		xg.ctx->vms.push_back(vm);
	}

	return true;
}

bool x_parse_context_data_block_information(void)
{
	x_vm_block_t		vmb;

	// number of blocks for virtualization
	if (!x_ctx_read_dword(&xg.ctx->num_blocks)) {
		return false;
	}

	if (!xg.ctx->num_blocks) {
		return false;
	}

	for (unsigned i = 0; i < xg.ctx->num_blocks; i++) {
		// vm owner
		if (!x_ctx_read_dword(&vmb.vm_id)) {
			return false;
		}

		if (vmb.vm_id >= xg.ctx->num_vms) {
			return false;
		}

		vmb.vm = &xg.ctx->vms[vmb.vm_id];

		// bytecode address
		if (!x_ctx_read_dword(&vmb.bc_rva)) {
			return false;
		}

		vmb.bc_va = xg.ctx->module_base + vmb.bc_rva;

		// invalid exec function by default
		vmb.exec_bc_function = ~0;

		// add a new block
		xg.ctx->vm_blocks.push_back(vmb);
	}

	return true;
}

bool x_parse_context_data_imports(void)
{
	for (unsigned i = 0; i < X_IMP_TOTAL; i++) {
		if (!x_ctx_read_dword(&xg.ctx->imports[i])) {
			return false;
		}
	}

	return true;
}

bool x_parse_context_data_data_addresses(void)
{
	return true;
}

bool x_initialize_assembler(void)
{
	x_cleanup_assembler();

	xg.code_gen_runtime = new StaticRuntime((void *) xg.ctx->new_code_va);
	xg.code_gen = new X86Assembler(xg.code_gen_runtime);

	return true;
}

bool x_initialize_temp_assembler(void)
{
	// cleanup assembler completely
	if (xg.temp_gen) {
		delete xg.temp_gen;
	}

	// cleanup assembler runtime
	if (xg.temp_gen) {
		delete xg.temp_runtime;
	}

	xg.temp_runtime = new StaticRuntime(NULL);
	xg.temp_gen = new X86Assembler(xg.temp_runtime);

	return true;
}

bool x_process_context(void)
{
	// initialize assembler
	if (!x_initialize_assembler()) {
		return false;
	}

	if (!x_initialize_temp_assembler()) {
		return false;
	}

	// initialize virtual machines
	x_init_virtual_machines();

	// register common functions
	x_register_generic_functions();

	// register vm functions
	x_register_vm_functions();

	// register junk instructions/operations
	x_register_junk_instructions();
	x_register_junk_operations();

	// register vm entry functions
	x_register_block_vm_entry_functions();

	// generate functions
	x_generate_functions();

	// split and mutate code
	{
		// initialize current pass
		x_init_shuffle_for_current_pass(0);

		// re-initialize assembler
		if (!x_initialize_assembler()) {
			x_exit_thread(X_ERR_PROCESS_CONTEXT_REINIT_ASSEMBLER);
		}

		// invalidate function addresses
		x_invalidate_function_addresses();

		// disassemble code block
		x_disassemble_block();

		// update all links in block
		x_update_disasm_owner_block();

		// analyze disassembly
		x_analyze_disassembly_in_block();

		// analyze relocations
		x_analyze_relocation_for_mutation();

		// make sure all instructions have been analyzed
		x_abort_if_instructions_not_analyzed();

		// clear relocation info
		x_clear_relocation_info();

		// split blocks
		x_split_blocks();

		// process splitted blocks
		x_process_block_parts();

		// verify function addrs
		x_verify_function_addresses();

		// make sure there's enough space for code
		xg.ctx->gen_data.resize(A.getCodeSize());

		// emit code
		A.relocCode(&xg.ctx->gen_data[0], xg.ctx->new_code_va);

		// clear existing block data
		x_clear_block();

		// set new block information
		//x_set_block(&xg.ctx->gen_data[0], &xg.ctx->gen_data[0] + A.getCodeSize());
	}

	// parse temporary relocations
	x_parse_temp_relocations();

	// set handler rvas
	x_generate_vm_handler_table();

	return true;
}

void x_cleanup_assembler(void)
{
	// cleanup assembler completely
	if (xg.code_gen) {
		delete xg.code_gen;
	}

	// cleanup assembler runtime
	if (xg.code_gen_runtime) {
		delete xg.code_gen_runtime;
	}
}

void x_cleanup_context(void)
{
	// cleanup assembler
	x_cleanup_assembler();

	// clear the entire context
	delete xg.ctx;
}

bool x_ctx_read_byte(UINT8 *db, bool deobfuscate)
{
	uint8_t		sb;

	if ((xg.ctx_data_offset + 1) > xg.ctx_data_len) {
		return false;
	}

	sb = *MakePtr(uint8_t *, xg.ctx_data, xg.ctx_data_offset);

	if (deobfuscate) {
		sb = x_ctx_de_obf_8(sb, xg.ctx_data_offset);
	}

	*db = sb;

	xg.ctx_data_offset += 1;

	return true;
}

bool x_ctx_read_word(UINT16 *dw, bool deobfuscate)
{
	uint16_t		sw;

	if ((xg.ctx_data_offset + 2) > xg.ctx_data_len) {
		return false;
	}

	sw = *MakePtr(uint16_t *, xg.ctx_data, xg.ctx_data_offset);

	if (deobfuscate) {
		sw = x_ctx_de_obf_16(sw, xg.ctx_data_offset);
	}

	*dw = sw;

	xg.ctx_data_offset += 2;

	return true;
}

bool x_ctx_read_dword(UINT32 *dd, bool deobfuscate)
{
	uint32_t		sd;

	if ((xg.ctx_data_offset + 4) > xg.ctx_data_len) {
		return false;
	}

	sd = *MakePtr(uint32_t *, xg.ctx_data, xg.ctx_data_offset);

	if (deobfuscate) {
		sd = x_ctx_de_obf_32(sd, xg.ctx_data_offset);
	}

	*dd = sd;

	xg.ctx_data_offset += 4;

	return true;
}

bool x_ctx_read_qword(UINT64 *dq, bool deobfuscate)
{
	union {
		uint64_t		q;
		struct {
			uint32_t	d1;
			uint32_t	d2;
		} s;
	} u;

	if (!x_ctx_read_dword(&u.s.d1, deobfuscate)) {
		return false;
	}

	if (!x_ctx_read_dword(&u.s.d2, deobfuscate)) {
		return false;
	}

	*dq = u.q;

	return true;
}

bool x_ctx_read_bool(bool *dst, bool deobfuscate)
{
	UINT8		val;

	if (!x_ctx_read_byte(&val, deobfuscate)) {
		return false;
	}

	*dst = (val != 0);

	return true;
}
