// x86obf - a code virtualizer/obfuscator for 32bit PE files
// Copyright (c) 2015 Tomislav Zubcic - chaplja.blogspot.com / x86obf@gmail.com
// $Id: x_vm_shared.h 652 2015-02-24 21:15:39Z tzubcic $

#pragma once

// common includes
#include "x_common.h"

// local headers
#include "x_vm_misc.h"

// ----------------------------------
// VM stack layout
// ----------------------------------

// vm stack size
#define X_VM_THREAD_DATA_SIZE		0x4000

// vm context size
#define X_VM_CONTEXT_SIZE			(X_VMR_TOTAL * 4)

// max number of return EIPs
#define X_VM_EIP_BACKUP_NUM			512

// bytecode decryption buffer
#define X_VM_BC_DEC_BUFF_SIZE		128

// return EIP buffer size
#define X_VM_EIP_BUFFER_SIZE		(X_VM_EIP_BACKUP_NUM * 4)

// base differenze size
#define X_VM_BASE_DIFF_SIZE			4

// random number size
#define X_VM_RAND_SIZE				4

// instruction execution
#define X_VM_INS_EXEC_SIZE			128

// orig instruction execution
#define X_VM_ORIG_EXEC_SIZE			128

// bytecode encryption key
#define X_VM_ENC_KEY_SIZE			4

// esp offset - static
#define X_VM_STACK_OFF_ESP			(X_VM_CONTEXT_SIZE + X_VM_BC_DEC_BUFF_SIZE + X_VM_EIP_BUFFER_SIZE + X_VM_BASE_DIFF_SIZE + X_VM_RAND_SIZE + X_VM_INS_EXEC_SIZE + X_VM_ORIG_EXEC_SIZE + X_VM_ENC_KEY_SIZE)

// ----------------------------------
// other constants
// ----------------------------------

// maximum number of virtual machines
#define X_VM_MAX_AMOUNT				8

// vm registers
#define X_VM_REG_NUM				32

// amount of fake vm instructions
#define X_VM_FAKE_INSTRUCTIONS		12

// instruction mnemonic length
#define X_VM_MNEMONIC_LEN			2

// max original instruction length
#define X_VM_MAX_ORIG_INSTR_LEN		16

// vm flags
#define X_VMF_SAVE_EFLAGS			(1 << 0)
#define X_VMF_ZERO					(1 << 1)

// vm condition codes
enum {
	X_CC_Z = 1,
	X_CC_NZ,

	// total number of condition codes
	X_CC_TOTAL
};

// required imported functions
enum {
	X_IMP_VIRTUALALLOC,
	X_IMP_TLSALLOC,
	X_IMP_TLSGETVALUE,
	X_IMP_TLSSETVALUE,
	X_IMP_EXITPROCESS,

	// total number of expected imports
	X_IMP_TOTAL
};

// vm state
enum {
	// status register
	X_VMR_N_EFLAGS,

	// directly accessible registers
	X_VMR_N_EAX,
	X_VMR_N_EBX,
	X_VMR_N_ECX,
	X_VMR_N_EDX,
	X_VMR_N_EDI,
	X_VMR_N_ESI,
	X_VMR_N_EBP,
	X_VMR_N_ESP,

	// vm registers
	X_VMR_VM_START,
	X_VMR_VM_END = X_VMR_VM_START + X_VM_REG_NUM - 1,

	// vm EIP
	X_VMR_VM_EIP,

	// flag saving toggle
	X_VMR_VM_FLAGS,

	// total number of registers
	X_VMR_TOTAL
};

// vm instructions
enum {
	// load register value/pointer
	X_VMI_MOV,						// mov R, R
	X_VMI_LEA,						// lea R, [R]

	// value assignments
	X_VMI_ASGN,						// asgn R, DWORD

	// read value
	X_VMI_RD,						// read BYTE, R, R

	// write value
	X_VMI_WR,						// wrb BYTE, R, R

	// arithmetic
	X_VMI_ADD,						// add BYTE, R, R
	X_VMI_SUB,						// sub BYTE, R, R

	// logical
	X_VMI_SHR,						// shr BYTE, R, R
	X_VMI_SHL,						// shl BYTE, R, R
	X_VMI_AND,						// and BYTE, R, R
	X_VMI_XOR,						// xor BYTE, R, R
	X_VMI_OR,						// or BYTE, R, R
	X_VMI_ROL,						// rol BYTE, R, R
	X_VMI_ROR,						// ror BYTE, R, R
	X_VMI_NOT,						// not BYTE, R

	// jumps
	X_VMI_JMP,						// jmp DWORD

	// leave vm
	X_VMI_LEAVE,					// leave

	// procedures
	X_VMI_CALL,						// call R

	// execute original instruction
	X_VMI_EXEC_BUF,					// eb R
	X_VMI_EXEC,						// exec BYTE, R

	// set/clear vm flags
	X_VMI_SETF,						// setf BYTE
	X_VMI_CLRF,						// clrf BYTE

	// relocation fix
	X_VMI_FIXRLC,					// fixrlc R

	// internal test/cmp
	X_VMI_VMTEST,					// vmtest R, R
	X_VMI_VMCMP,					// vmcmp R, R

	// total number of instructions
	X_VMI_TOTAL
};

// vm implementation functions
enum {
	// generic vm functions
	X_VM_FUNC_ENTER,
	X_VM_FUNC_LEAVE,
	X_VM_FUNC_LEAVE_CALL_RET,
	X_VM_FUNC_LEAVE_ORIG_RET,
	X_VM_FUNC_STORE_EIP,
	X_VM_FUNC_GET_EIP,
	X_VM_FUNC_INIT_THREAD_DATA,
	X_VM_FUNC_GET_THREAD_DATA,

	// vm instructions
	X_VM_FUNC_I_MOV,
	X_VM_FUNC_I_LEA,
	X_VM_FUNC_I_ASGN,
	X_VM_FUNC_I_RD,
	X_VM_FUNC_I_WR,
	X_VM_FUNC_I_ADD,
	X_VM_FUNC_I_SUB,
	X_VM_FUNC_I_SHR,
	X_VM_FUNC_I_SHL,
	X_VM_FUNC_I_AND,
	X_VM_FUNC_I_XOR,
	X_VM_FUNC_I_OR,
	X_VM_FUNC_I_ROL,
	X_VM_FUNC_I_ROR,
	X_VM_FUNC_I_NOT,
	X_VM_FUNC_I_JMP,
	X_VM_FUNC_I_LEAVE,
	X_VM_FUNC_I_CALL,
	X_VM_FUNC_I_EXEC_BUF,
	X_VM_FUNC_I_EXEC,
	X_VM_FUNC_I_SETF,
	X_VM_FUNC_I_CLRF,
	X_VM_FUNC_I_FIXRLC,
	X_VM_FUNC_I_VMTEST,
	X_VM_FUNC_I_VMCMP,

	// total number of vm functions
	X_VM_FUNC_TOTAL
};

// vm instruction operand types
enum {
	X_VMI_OP_NONE,
	X_VMI_OP_REG,
	X_VMI_OP_BYTE,
	X_VMI_OP_DWORD,
	X_VMI_OP_REL_DWORD,
	X_VMI_OP_SIZE,

	// total number of operand types
	X_VMI_OP_TOTAL
};

typedef struct {
	UINT32				id;
	UINT32				op_types[4];
	unsigned			func_id;
} x_vm_ins_info_t;

// get instruction info
unsigned x_get_vm_operand_len(UINT32);
x_vm_ins_info_t *x_get_vm_ins_info(UINT32);
UINT32 x_get_vm_ins_size(UINT32);

// get operand info
UINT32 x_get_vm_op_offset(UINT32, UINT32);
UINT32 x_get_vm_op_size(UINT32, UINT32);

// encode values in vm
UINT8 x_encode_byte(UINT32, UINT32, unsigned);
UINT32 x_encode_dword(UINT32, UINT32, unsigned);
